# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/protocols/base.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bopenfl/protocols/base.proto\"+\n\nModelProto\x12\x1d\n\x07tensors\x18\x01 \x03(\x0b\x32\x0c.NamedTensor\"\xa3\x01\n\x0bNamedTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x10\n\x08lossless\x18\x03 \x01(\x08\x12\x0e\n\x06report\x18\x04 \x01(\x08\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12,\n\x14transformer_metadata\x18\x06 \x03(\x0b\x32\x0e.MetadataProto\x12\x12\n\ndata_bytes\x18\x07 \x01(\x0c\"\x9d\x01\n\rMetadataProto\x12\x34\n\x0cint_to_float\x18\x01 \x03(\x0b\x32\x1e.MetadataProto.IntToFloatEntry\x12\x10\n\x08int_list\x18\x02 \x03(\x05\x12\x11\n\tbool_list\x18\x03 \x03(\x08\x1a\x31\n\x0fIntToFloatEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\"+\n\nDataStream\x12\x0c\n\x04size\x18\x01 \x01(\r\x12\x0f\n\x07npbytes\x18\x02 \x01(\x0c\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'openfl.protocols.base_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_METADATAPROTO_INTTOFLOATENTRY']._loaded_options = None
  _globals['_METADATAPROTO_INTTOFLOATENTRY']._serialized_options = b'8\001'
  _globals['_MODELPROTO']._serialized_start=31
  _globals['_MODELPROTO']._serialized_end=74
  _globals['_NAMEDTENSOR']._serialized_start=77
  _globals['_NAMEDTENSOR']._serialized_end=240
  _globals['_METADATAPROTO']._serialized_start=243
  _globals['_METADATAPROTO']._serialized_end=400
  _globals['_METADATAPROTO_INTTOFLOATENTRY']._serialized_start=351
  _globals['_METADATAPROTO_INTTOFLOATENTRY']._serialized_end=400
  _globals['_DATASTREAM']._serialized_start=402
  _globals['_DATASTREAM']._serialized_end=445
# @@protoc_insertion_point(module_scope)
