# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from openfl.interface.aggregation_functions.experimental.privileged_aggregation import (
    PrivilegedAggregationFunction,
)
