# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/protocols/aggregator.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!openfl/protocols/aggregator.proto\x12\x11openfl.aggregator\x1a\x1bopenfl/protocols/base.proto\"o\n\rMessageHeader\x12\x0e\n\x06sender\x18\x01 \x01(\t\x12\x10\n\x08receiver\x18\x02 \x01(\t\x12\x17\n\x0f\x66\x65\x64\x65ration_uuid\x18\x03 \x01(\t\x12#\n\x1bsingle_col_cert_common_name\x18\x04 \x01(\t\"?\n\x0bPingRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\"@\n\x0cPingResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\"C\n\x0fGetTasksRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\"S\n\x04Task\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rfunction_name\x18\x02 \x01(\t\x12\x11\n\ttask_type\x18\x03 \x01(\t\x12\x13\n\x0b\x61pply_local\x18\x04 \x01(\x08\"\xa4\x01\n\x10GetTasksResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12&\n\x05tasks\x18\x03 \x03(\x0b\x32\x17.openfl.aggregator.Task\x12\x12\n\nsleep_time\x18\x04 \x01(\x05\x12\x0c\n\x04quit\x18\x05 \x01(\x08\"\xb1\x01\n\x1aGetAggregatedTensorRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x13\n\x0btensor_name\x18\x02 \x01(\t\x12\x14\n\x0cround_number\x18\x03 \x01(\x05\x12\x0e\n\x06report\x18\x04 \x01(\x08\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12\x18\n\x10require_lossless\x18\x06 \x01(\x08\"\x83\x01\n\x1bGetAggregatedTensorResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x1c\n\x06tensor\x18\x03 \x01(\x0b\x32\x0c.NamedTensor\"\x9a\x01\n\x0bTaskResults\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x11\n\ttask_name\x18\x03 \x01(\t\x12\x11\n\tdata_size\x18\x04 \x01(\x05\x12\x1d\n\x07tensors\x18\x05 \x03(\x0b\x32\x0c.NamedTensor\"P\n\x1cSendLocalTaskResultsResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\"\xd4\x01\n\x0eInteropMessage\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x1c\n\x07message\x18\x02 \x01(\x0b\x32\x0b.DataStream\x12\x41\n\x08metadata\x18\x03 \x03(\x0b\x32/.openfl.aggregator.InteropMessage.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32\xd8\x03\n\nAggregator\x12I\n\x04Ping\x12\x1e.openfl.aggregator.PingRequest\x1a\x1f.openfl.aggregator.PingResponse\"\x00\x12U\n\x08GetTasks\x12\".openfl.aggregator.GetTasksRequest\x1a#.openfl.aggregator.GetTasksResponse\"\x00\x12v\n\x13GetAggregatedTensor\x12-.openfl.aggregator.GetAggregatedTensorRequest\x1a..openfl.aggregator.GetAggregatedTensorResponse\"\x00\x12X\n\x14SendLocalTaskResults\x12\x0b.DataStream\x1a/.openfl.aggregator.SendLocalTaskResultsResponse\"\x00(\x01\x12V\n\x0cInteropRelay\x12!.openfl.aggregator.InteropMessage\x1a!.openfl.aggregator.InteropMessage\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'openfl.protocols.aggregator_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_INTEROPMESSAGE_METADATAENTRY']._loaded_options = None
  _globals['_INTEROPMESSAGE_METADATAENTRY']._serialized_options = b'8\001'
  _globals['_MESSAGEHEADER']._serialized_start=85
  _globals['_MESSAGEHEADER']._serialized_end=196
  _globals['_PINGREQUEST']._serialized_start=198
  _globals['_PINGREQUEST']._serialized_end=261
  _globals['_PINGRESPONSE']._serialized_start=263
  _globals['_PINGRESPONSE']._serialized_end=327
  _globals['_GETTASKSREQUEST']._serialized_start=329
  _globals['_GETTASKSREQUEST']._serialized_end=396
  _globals['_TASK']._serialized_start=398
  _globals['_TASK']._serialized_end=481
  _globals['_GETTASKSRESPONSE']._serialized_start=484
  _globals['_GETTASKSRESPONSE']._serialized_end=648
  _globals['_GETAGGREGATEDTENSORREQUEST']._serialized_start=651
  _globals['_GETAGGREGATEDTENSORREQUEST']._serialized_end=828
  _globals['_GETAGGREGATEDTENSORRESPONSE']._serialized_start=831
  _globals['_GETAGGREGATEDTENSORRESPONSE']._serialized_end=962
  _globals['_TASKRESULTS']._serialized_start=965
  _globals['_TASKRESULTS']._serialized_end=1119
  _globals['_SENDLOCALTASKRESULTSRESPONSE']._serialized_start=1121
  _globals['_SENDLOCALTASKRESULTSRESPONSE']._serialized_end=1201
  _globals['_INTEROPMESSAGE']._serialized_start=1204
  _globals['_INTEROPMESSAGE']._serialized_end=1416
  _globals['_INTEROPMESSAGE_METADATAENTRY']._serialized_start=1369
  _globals['_INTEROPMESSAGE_METADATAENTRY']._serialized_end=1416
  _globals['_AGGREGATOR']._serialized_start=1419
  _globals['_AGGREGATOR']._serialized_end=1891
# @@protoc_insertion_point(module_scope)
